import uproot
import os

import pandas as pd

from tools.fail_to_pass import fail_to_pass
import utils_noroot    as utnr

log=utnr.getLogger(__name__)

def read_root(year: str, trig: str, channel: str, sel, pbdt=False) -> pd.DataFrame:
    """
    Function to read root for the hadron mis-ID study.
    input : year, trigger category of samples you want.
            sel: additional selection.
    output: pandas.DataFrame that contents events.
    """
    force_remake_samples = True

    version = 'v10.21p2' if channel in [ 'bpkkk', 'bpkpipi' ] else 'v10.18is'
    
    cached_path = '/afs/ihep.ac.cn/users/x/xzh6313/xzh/RKst/dev_zhihao/differential-crosscheck-fits/data/tools/apply_selection'

    if pbdt == False:
        roots_path = f'{cached_path}/noPID/high/{channel}/{version}/{year}_{trig}'
    elif pbdt == True and channel in ['cmb', 'data']:
        roots_path = f'{cached_path}/noPID/high_allsel/{channel}/{version}/{year}_{trig}'
    else:
        raise KeyError(f'For channel {channel}, full selected sample is not supported now.')

    desired_branches  = [ 
                          "B_M",
                          "L1_PIDe", "L1_P", "L1_PT", "L1_ETA", "L1_ProbNNk",
                          "L2_PIDe", "L2_P", "L2_PT", "L2_ETA", "L2_ProbNNk",
                          "H_PIDe" , "H_ProbNNk"    , "Polarity"
                        ]
    
    merged_root_file = f'{roots_path}/merged_10.root'

    if os.path.exists(merged_root_file) and force_remake_samples == False:
            log.debug('Washed&Merged samples exists, loading...')
    else:

        with uproot.recreate(merged_root_file) as merged_file:
            log.debug('Washing and Merging input samples...')
            merged_tree = []

            for i in range(10):
                tem_rt_name = f'{roots_path}/{i}_10.root'
            
                with uproot.open(f'{tem_rt_name}:{trig}') as tree:
                    tree_sel = tree.arrays(desired_branches, library='pd')
                merged_tree.append(tree_sel)

            merged_tree = pd.concat(merged_tree)
            merged_file[trig] = merged_tree

    with uproot.open(f'{merged_root_file}:{trig}') as tree:
        df = tree.arrays(library='pd')
    
    if sel is not None:
        df = df.query(sel)

    return df

def read_weight(year: str, mag: str, particle: str):
    """
    Function to read weight maps.
    input : year, magnet and particle of samples you want.
    output: Hist that contents weight.
    """
    file_path = f"/afs/ihep.ac.cn/users/x/xzh6313/xzh/RK/dev_zhihao/hadron-mis-id/storage/{year}_{mag}_{particle}_w_map.root"
    with uproot.open(file_path) as f:
        hist = f[f"{particle}_DLLe>3_All"].to_hist()
    return hist

def add_kinematic_weight(df):
    ...

def add_pid_effciency_weight(df):
    ...

def get_weighted_sample(kind, year, trig, channel, pbdt=False, wgt=True)-> tuple[pd.DataFrame, tuple[pd.DataFrame,pd.DataFrame,pd.DataFrame]]:
    """
    output: pp_df, ( fp_df, pf_df, ff_df )
    sample isself is not weighted.
    use df['w'] to apply weight manually.
    """
    public_sel = 'H_ProbNNk>0.2&H_PIDe<0&L1_PIDe>-2&L2_PIDe>-2'

    if kind == 'K':
        sel = f'L1_ProbNNk>0.1&L2_ProbNNk>0.1&{public_sel}'
    elif kind == 'Pi':
        sel = f'L1_ProbNNk<0.1&L2_ProbNNk<0.1&{public_sel}'
    elif kind == 'H':
        sel = f'{public_sel}'
    else:
        raise ValueError(f'Kind {kind} is not supported now.')

    data = read_root(year, trig, channel, sel, pbdt=pbdt)
    if wgt == False:
        return data
    # if channel in [ 'bpkkk', 'bpkpipi' ]:
    #     # For sample bpKKk and bpkpipi
    #     # Kinematic weight should be added
    #     # PID efficiencies calculated by PIDCalib added instead of PID cut.

    #     data = add_kinematic_weight(data)
    #     data_K, data_Pi = add_pid_effciency_weight(data)

    #     def replace_probNNk(df, pacticle, value):
    #         row = f'{pacticle}_probNNk'
    #         df[row] = df.apply(lambda x: value)

    #         return df

    #     data_K = replace_probNNk(data, 'L1', 1)
    #     data_K = replace_probNNk(data, 'L2', 1)

    #     data_Pi = replace_probNNk(data, 'L1', 0)
    #     data_Pi = replace_probNNk(data, 'L2', 0)

    #     data = pd.concat( [data_K, data_Pi] )

    data_MD = data.query("Polarity==-1")
    data_MU = data.query("Polarity==1")

    MD_maps = { "K": read_weight(year, "MagDown", "K"), "Pi": read_weight(year, "MagDown", "Pi") }
    MU_maps = { "K": read_weight(year, "MagUp", "K")  , "Pi": read_weight(year, "MagUp", "Pi") }

    selector = lambda row, p: row[f"{p}_PIDe"] > 3
    map_selector = lambda row, p, maps: maps["K"] if row[f"{p}_ProbNNk"] > 0.1 else maps["Pi"]

    ftp = fail_to_pass(data_MD, "L1", "L2", selector, MD_maps, map_selector, ["PT", "ETA"])
    data_MD    = ftp.get_weighted_sample()
    data_MD_FP = ftp.df_fp
    data_MD_PF = ftp.df_pf
    data_MD_FF = ftp.df_ff

    ftp = fail_to_pass(data_MU, "L1", "L2", selector, MU_maps, map_selector, ["PT", "ETA"])
    data_MU    = ftp.get_weighted_sample()
    data_MU_FP = ftp.df_fp
    data_MU_PF = ftp.df_pf
    data_MU_FF = ftp.df_ff

    data    = pd.concat( [data_MU   , data_MD   ] )
    data_fp = pd.concat( [data_MU_FP, data_MD_FP] )
    data_pf = pd.concat( [data_MU_PF, data_MD_PF] )
    data_ff = pd.concat( [data_MU_FF, data_MD_FF] )

    if len(data_ff) != 0:
        data_ff['w'] = -1 * data_ff['w']

    return data, (data_fp, data_pf, data_ff)



from misID_tools.samples import get_weighted_sample

import matplotlib.pyplot as plt
import numpy as np
import mplhep

def plot_data(df, name):
    bins = 80

    limits = (4000, 6500)

    counts, bin_edges = np.histogram(df['B_M'], bins, range=limits)

    # Substitute plt.hist()
    mplhep.histplot((counts, bin_edges), yerr=True, color='black', histtype='errorbar')

    plt.xlabel('B_M')

    plt.savefig(f'./output/{name}.png')
    plt.close('all')

    return 1

def test_samples():
    _, (_, _, Ka_ff) = get_weighted_sample(kind='K' , year='2018', trig='ETOS', channel='data')
    _, (_, _, Pi_ff) = get_weighted_sample(kind='Pi', year='2018', trig='ETOS', channel='data')
    Ha_pp, _ = get_weighted_sample(kind='H' , year='2018', trig='ETOS', channel='data')

    assert plot_data(Ka_ff, 'Kaon_failfail')
    assert plot_data(Pi_ff, 'Pion_failfail')
    assert plot_data(Ha_pp, 'All_passpass')

     


from setuptools import setup, find_packages

import glob

setup(
        name            ='rk_hadmisid_study',
        version         ='0.0.5',
        author          ='xuzh',
        description     ='All tools and scripts used in the study of the hadron mis-ID.',
        packages        = find_packages(),
        package_data    = {'misID_data' : ['model/*/*.csv', 'model/*/*.json']},
        install_requires= open('requirements.txt').read()
        )



from misID_tools.zmodel import misID_real_model_builder
from zutils.plot import plot as zfp

import zutils.utils as zut
import utils_noroot as utnr
import matplotlib.pyplot as plt
import zfit
import os

class data:
    ver = 'v1'              
    obs = zfit.Space('B_M', limits=(4500, 6500))
    out_dir = 'output/zmodel' 

    os.makedirs(out_dir, exist_ok=True) 

def test_real_model_IO():
    misID_builder = misID_real_model_builder('misID_1', version=data.ver, obs=data.obs)

    try:
        misID_builder.load_model()
        model = misID_builder.build_model()
    except:
        assert False

    zut.print_pdf(model, txt_path=f'{data.out_dir}/pdf_IO.txt') 
    
def test_real_model_function():
    misID_builder = misID_real_model_builder('misID_2', version=data.ver, obs=data.obs)

    misID_builder.load_model()
    # misID_builder.fix_mode = 'all' # Default option even if you don't sepcify.
    model = misID_builder.build_model()

    try:
        N = sum(misID_builder.params['yld'][1:])
        sampler = model.create_sampler(n=N,fixed_params=True)

        plotter = zfp(data=sampler, model=model)
        plotter.plot(nbins=50, plot_range=(4500, 6500))
        
        plt.savefig(f'{data.out_dir}/plot_misID.png', bbox_inches='tight')
        plt.close('all')
    except:
        assert False

    zut.print_pdf(model, txt_path=f'{data.out_dir}/pdf_function.txt') 

def test_real_model_fix_ratio():
    misID_builder = misID_real_model_builder('misID_3', version=data.ver, obs=data.obs)

    misID_builder.load_model()
    misID_builder.fix_mode = 'ratio' # Must be set before call "build_model()"
    model = misID_builder.build_model()

    try:
        N = 10000
        sampler = model.create_sampler(n=N,fixed_params=True)

        model.fit_to(sampler)

        plotter = zfp(data=sampler, model=model)
        plotter.plot(nbins=50, plot_range=(4500, 6500))
        
        plt.savefig(f'{data.out_dir}/plot_misID_ratio_mode.png', bbox_inches='tight')
        plt.close('all')
    except:
        assert False

    zut.print_pdf(model, txt_path=f'{data.out_dir}/pdf_fix_ratio.txt') 

if __name__ == '__main__':
    test_real_model_IO()
    test_real_model_function()
    test_real_model_fix_ratio()




import zfit
import os
from zfit import z
import tensorflow as tf
import pandas as pd

from importlib.resources import files

from zmodel import diff_base_model
from zutils.pdf import SUJohnson
import utils_noroot as utnr


log = utnr.getLogger(__name__)
class Empirical(zfit.pdf.BasePDF):

    def __init__(self, m0, b1, b2, a, obs, extended=None, norm=None, name=None):
        params = { 
                   'm0' : m0,  
                   'b1' : b1,
                   'b2' : b2, 
                   'a'  : a ,
                 }
        super().__init__(obs=obs, params=params, extended=extended, norm=norm, name=name)

    def _unnormalized_pdf(self, x):
        x = z.unstack_x(x)
        # ma = self.params['ma']
        # mb = self.params['mb']
        m0 = self.params['m0']
        b1 = self.params['b1']
        b2 = self.params['b2']
        a  = self.params['a']

        dx = x - m0
        y = dx**a * z.exp( -dx*b1 + dx*dx*b2*0 )
        # n = ( ( x**2 - ma**2 - mb**2 )**2 - 4*ma**2*mb**2 )**0.5
        # dn= 2*x

        cond = tf.greater(x, m0)

        return tf.where(cond, y, 0)

class misID_base_model(diff_base_model):
    def __init__(self, name, params: dict, samples: dict):
        self.name     = name
        self._params  = params
        self._samples = samples
        self._obs     = zfit.Space('B_M', limits=(4500, 6500))

        self._model_built = False

class My_SUJohnson(misID_base_model):
    def _initial_model(self):
        l_param_phsp = self.params['SU']

        if hasattr(self, 'fix') and self.fix == False:
            mu           = zfit.param.Parameter(f"{self.name}_mu",  l_param_phsp[0], 4000, 5000)
            lm           = zfit.param.Parameter(f"{self.name}_lm",  l_param_phsp[1], 0   , 100 )
            gm           = zfit.param.Parameter(f"{self.name}_gm",  l_param_phsp[2],-20  , 20  )
            dt           = zfit.param.Parameter(f"{self.name}_dt",  l_param_phsp[3], 0   , 10  )
        else:
            mu           = zfit.param.ConstantParameter(f"{self.name}_mu",  l_param_phsp[0])
            lm           = zfit.param.ConstantParameter(f"{self.name}_lm",  l_param_phsp[1])
            gm           = zfit.param.ConstantParameter(f"{self.name}_gm",  l_param_phsp[2])
            dt           = zfit.param.ConstantParameter(f"{self.name}_dt",  l_param_phsp[3])

        y_SUJs        = zfit.Parameter(f"{self.name}_NSUJs"    , self.params['yld'][0], 0, 1000)

        pdf_SUJohnson = SUJohnson(obs=self._obs, mu=mu, lm=lm, gamma=gm, delta=dt, name='SUJohnson')
        pdf_SUJohnson.set_yield(y_SUJs)

        self._total_model = pdf_SUJohnson
        self._model_built = True

class misID_pp_model_builder(misID_base_model):
    """
    For mimic pass-pass sample fit,  components needed:
    - Combinatorial background. Described by as nominal model, SU johnson.
    - B0/+ -> K*0/+ e+ e-. Described by GridKDE from simulation sample.
    - B+ -> K+ e+ e-. Marker as signal, described by GridKDE from simulation sample.
    - B+ -> K+ K+ K-/K+ pi+ pi-. Described by GridKDE from simulation sample.
    - Other mis-ID components. Described by phase-space model.
    """
    def _initial_model(self):

        def key_to_zfit_data(label, wgt=True): 
            data_df   = self._samples[label]
            if wgt == True:
                data_zfit = zfit.Data.from_pandas(data_df['B_M'], obs=self._obs, weights=data_df['w'])
            else:
                data_zfit = zfit.Data.from_pandas(data_df['B_M'], obs=self._obs)
            return data_zfit
        
        total_pdf = []

        # Define singal pdf
        data_sig     = key_to_zfit_data('sign', wgt=False)
        y_sig        = zfit.param.Parameter(f"{self.name}_Nsign"     , self.params['yld'][0])
        pdf_sig      = zfit.pdf.KDE1DimGrid(data_sig, name='sign', obs=self._obs)
        pdf_sig.set_yield(y_sig)
        total_pdf.append(pdf_sig)

        # Define B0 -> K*0 e+ e-
        data_prec_bd = key_to_zfit_data('bdks', wgt=False)
        y_prec_bd    = zfit.Parameter(f"{self.name}_Nbdks" , self.params['yld'][1], 0, 1e2)
        pdf_prec_bd  = zfit.pdf.KDE1DimGrid(data_prec_bd, name='bdks', obs=self._obs)
        pdf_prec_bd.set_yield(y_prec_bd)
        total_pdf.append(pdf_prec_bd)

        # Define B+ -> K*+ e+ e-
        data_prec_bp = key_to_zfit_data('bpks', wgt=False)
        y_prec_bp    = zfit.Parameter(f"{self.name}_Nbpks" , self.params['yld'][2], 0, 1e2)
        pdf_prec_bp  = zfit.pdf.KDE1DimGrid(data_prec_bp, name='bpks', obs=self._obs)
        pdf_prec_bp.set_yield(y_prec_bp)
        total_pdf.append(pdf_prec_bp)

        # # Define B+ -> K+ K+ K-
        data_bpkkk   = key_to_zfit_data('bpkkk', wgt=False)
        y_bpkkk      = zfit.param.ConstantParameter(f"{self.name}_Nbpkkk"    , self.params['yld'][3])
        pdf_bpkkk    = zfit.pdf.KDE1DimGrid(data_bpkkk, name='bpkkk', obs=self._obs)
        pdf_bpkkk.set_yield(y_bpkkk)
        total_pdf.append(pdf_bpkkk)

        # Define B+ -> K+ pi+ pi-
        data_bpkpipi = key_to_zfit_data('bpkpipi', wgt=False)
        y_bpkpipi    = zfit.param.ConstantParameter(f"{self.name}_Nbpkpipi"  , self.params['yld'][4])
        pdf_bpkpipi  = zfit.pdf.KDE1DimGrid(data_bpkpipi, name='bpkpipi', obs=self._obs)
        pdf_bpkpipi.set_yield(y_bpkpipi)
        total_pdf.append(pdf_bpkpipi)

        # Define comb. and others
        l_param_phsp = self.params['phsp']
        m0           = zfit.param.ConstantParameter(f"{self.name}_m0" ,  l_param_phsp[0])

        b1           = zfit.param.Parameter(f"{self.name}_b1" ,  l_param_phsp[1], 0, 10e-3)
        b2           = zfit.param.ConstantParameter(f"{self.name}_b2" ,  l_param_phsp[2])

        a            = zfit.param.ConstantParameter(f"{self.name}_a"  ,  l_param_phsp[3])

        y_phsp       = zfit.Parameter(f"{self.name}_Nphsp"      , self.params['yld'][5], 0, 3e2)
        pdf_phsp     = Empirical(m0, b1, b2, a, name='other mis-ID', obs=self._obs)
        pdf_phsp.set_yield(y_phsp)
        total_pdf.append(pdf_phsp)

        l_param_SU = self.params['SU']
        mu           = zfit.param.ConstantParameter(f"{self.name}_mu",  l_param_SU[0])
        lm           = zfit.param.ConstantParameter(f"{self.name}_lm",  l_param_SU[1])
        gm           = zfit.param.ConstantParameter(f"{self.name}_gm",  l_param_SU[2])
        dt           = zfit.param.ConstantParameter(f"{self.name}_dt",  l_param_SU[3])

        y_SUJs        = zfit.param.ConstantParameter(f"{self.name}_NSUJs", self.params['yld'][6])

        pdf_SUJohnson = SUJohnson(obs=self._obs, mu=mu, lm=lm, gamma=gm, delta=dt, name='Combinatorial')
        pdf_SUJohnson.set_yield(y_SUJs)
        total_pdf.append(pdf_SUJohnson)

        # self._total_model = pdf_phsp
        self._total_model = zfit.pdf.SumPDF( total_pdf, name=f"{self.name}_totpdf" )
        self._model_built = True

class misID_ff_model_builder(misID_base_model):
    """
    input: name, params: {'phsp':[2], 'yld':[6] }, samples: {<key> : <sample> }[5]

    For fail-fail sample fit,  components needed:
    - B0/+ -> K*0/+ e+ e-. Described by GridKDE from simulation sample.
    - B+ -> K+ e+ e-. Marker as signal, described by GridKDE from simulation sample.
    - B+ -> K+ K+ K-/K+ pi+ pi-. Described by GridKDE from simulation sample.
    - Other mis-ID components and Combinatorial background. Described by phase-space model.
    """
    def _initial_model(self):

        def key_to_zfit_data(label, wgt=True): 
            data_df   = self._samples[label]
            if wgt == True:
                data_zfit = zfit.Data.from_pandas(data_df['B_M'], obs=self._obs, weights=data_df['w'])
            else:
                data_zfit = zfit.Data.from_pandas(data_df['B_M'], obs=self._obs)
            return data_zfit
        
        total_pdf = []

        # Define singal pdf
        data_sig     = key_to_zfit_data('sign', wgt=False)
        y_sig        = zfit.Parameter(f"{self.name}_Nsign"     , self.params['yld'][0], 0, 1e4)
        pdf_sig      = zfit.pdf.KDE1DimGrid(data_sig, name='sign', obs=self._obs)
        pdf_sig.set_yield(y_sig)
        total_pdf.append(pdf_sig)

        # Define B0 -> K*0 e+ e-
        data_prec_bd = key_to_zfit_data('bdks', wgt=False)
        y_prec_bd    = zfit.Parameter(f"{self.name}_Nbdks" , self.params['yld'][1], 0, 1e4)
        pdf_prec_bd  = zfit.pdf.KDE1DimGrid(data_prec_bd, name='bdks', obs=self._obs)
        pdf_prec_bd.set_yield(y_prec_bd)
        total_pdf.append(pdf_prec_bd)

        # Define B+ -> K*+ e+ e-
        data_prec_bp = key_to_zfit_data('bpks', wgt=False)
        y_prec_bp    = zfit.Parameter(f"{self.name}_Nbpks" , self.params['yld'][2], 0, 1e4)
        pdf_prec_bp  = zfit.pdf.KDE1DimGrid(data_prec_bp, name='bpks', obs=self._obs)
        pdf_prec_bp.set_yield(y_prec_bp)
        total_pdf.append(pdf_prec_bp)

        # # Define B+ -> K+ K+ K-
        data_bpkkk   = key_to_zfit_data('bpkkk', wgt=False)
        y_bpkkk      = zfit.Parameter(f"{self.name}_Nbpkkk"    , self.params['yld'][3], 0, 1e4)
        pdf_bpkkk    = zfit.pdf.KDE1DimGrid(data_bpkkk, name='bpkkk', obs=self._obs)
        pdf_bpkkk.set_yield(y_bpkkk)
        total_pdf.append(pdf_bpkkk)

        # Define B+ -> K+ pi+ pi-
        data_bpkpipi = key_to_zfit_data('bpkpipi', wgt=False)
        y_bpkpipi    = zfit.Parameter(f"{self.name}_Nbpkpipi"  , self.params['yld'][4], 0, 1e4)
        pdf_bpkpipi  = zfit.pdf.KDE1DimGrid(data_bpkpipi, name='bpkpipi', obs=self._obs)
        pdf_bpkpipi.set_yield(y_bpkpipi)
        total_pdf.append(pdf_bpkpipi)

        # Define comb. and others
        l_param_phsp = self.params['phsp']
        m0           = zfit.param.Parameter(f"{self.name}_m0" ,  l_param_phsp[0], l_param_phsp[0] - 0.05*l_param_phsp[0], l_param_phsp[0] + 0.05*l_param_phsp[0])

        b1           = zfit.param.Parameter(f"{self.name}_b1" ,  l_param_phsp[1])
        b2           = zfit.param.Parameter(f"{self.name}_b2" ,  l_param_phsp[2])

        y_phsp       = zfit.Parameter(f"{self.name}_Nphsp"      , self.params['yld'][5], 0, 1e4)
        pdf_phsp     = Empirical(m0, b1, b2, name='other mis-ID & cmb', obs=self._obs)
        pdf_phsp.set_yield(y_phsp)
        total_pdf.append(pdf_phsp)

        # self._total_model = pdf_phsp
        self._total_model = zfit.pdf.SumPDF( total_pdf, name=f"{self.name}_totpdf" )
        self._model_built = True

class misID_real_model_builder(misID_base_model):
    """
    The pdf builder for the total mis-ID pdf.
    
    Usage:
    ```
    misID_builder = misID_real_model_builder(name)
    misID_builder.load_model(path)
    
    model = misID_builder.build_model() # you got a zfit.pdf.SumPDF as model.

    ```
    """

    def __init__(self, name=None, version=None, obs=None):
        self.name     = name
        self._vers    = version
        self._obs     = obs 

        self._pars_dir    = None
        self._model_built = False
        self._initialized = False

    def _initialize(self):
        if self._initialized:
            return

        if self._obs is None:
            log.error(f'Observable is missing')
            raise

        self._pars_dir=files('misID_data').joinpath(f'model/{self._vers}')
        if not os.path.isdir(self._pars_dir):
            log.error(f'Missing directory with data: {self._pars_dir}')
            raise FileNotFoundError

        self._initialized = True

    def _initial_model_by_fix_mode(self, fix_mode):
        def key_to_zfit_data(label, wgt=True): 
            data_df   = self._samples[label]
            if wgt == True:
                data_zfit = zfit.Data.from_pandas(data_df['B_M'], obs=self._obs, weights=data_df['w'])
            else:
                data_zfit = zfit.Data.from_pandas(data_df['B_M'], obs=self._obs)
            return data_zfit
        
        total_pdf = []

        # # Define B+ -> K+ K+ K-
        data_bpkkk   = key_to_zfit_data('bpkkk', wgt=False)
        pdf_bpkkk    = zfit.pdf.KDE1DimGrid(data_bpkkk, name='bpkkk', obs=self._obs)

        # Define B+ -> K+ pi+ pi-
        data_bpkpipi = key_to_zfit_data('bpkpipi', wgt=False)
        pdf_bpkpipi  = zfit.pdf.KDE1DimGrid(data_bpkpipi, name='bpkpipi', obs=self._obs)

        # Define others
        # l_param_phsp = self.params['phsp']
        # m0           = zfit.param.ConstantParameter(f"{self.name}_m0" ,  l_param_phsp[0])

        # b1           = zfit.param.ConstantParameter(f"{self.name}_b1" ,  l_param_phsp[1])
        # b2           = zfit.param.ConstantParameter(f"{self.name}_b2" ,  l_param_phsp[2])
        # a            = zfit.param.ConstantParameter(f"{self.name}_a"  ,  l_param_phsp[3])

        # pdf_phsp     = Empirical(m0, b1, b2, a, name='other mis-ID', obs=self._obs)

        if fix_mode == 'all':
            y_bpkkk   = zfit.param.ConstantParameter(f"{self.name}_Nbpkkk"    , self.params['yld'][0])
            y_bpkpipi = zfit.param.ConstantParameter(f"{self.name}_Nbpkpipi"  , self.params['yld'][1])
            # y_phsp    = zfit.param.ConstantParameter(f"{self.name}_Nphsp"      , self.params['yld'][2])
        elif fix_mode == 'ratio':
            y_total   = zfit.param.Parameter(f"{self.name}_NallmisID"       , self.params['yld'][0]+self.params['yld'][1], 0, 1e8)

            r_bpkkk   = self.params['yld'][0]/(self.params['yld'][0]+self.params['yld'][1])
            y_bpkkk   = zfit.param.ComposedParameter(f"{self.name}_Nbpkkk"  , lambda params: params['a']*r_bpkkk, params={'a':y_total})

            r_bpkpipi = self.params['yld'][1]/(self.params['yld'][0]+self.params['yld'][1])
            y_bpkpipi = zfit.param.ComposedParameter(f"{self.name}_Nbpkpipi", lambda params: params['a']*r_bpkpipi, params={'a':y_total})

            # r_phsp    = self.params['yld'][2]/sum(self.params['yld'])
            # y_phsp    = zfit.param.ComposedParameter(f"{self.name}_Nphsp"   , lambda: params[0]*r_phsp, params=[y_total])
        else:
            raise ValueError(f'fix mode of {fix_mode} is not supported.')

        pdf_bpkkk.set_yield(y_bpkkk)
        pdf_bpkpipi.set_yield(y_bpkpipi)
        # pdf_phsp.set_yield(y_phsp)

        total_pdf.append(pdf_bpkkk)
        total_pdf.append(pdf_bpkpipi)
        # total_pdf.append(pdf_phsp)

        # self._total_model = pdf_phsp
        return zfit.pdf.SumPDF( total_pdf, name=f"{self.name}_totpdf" )

    def _initial_model(self):
        if hasattr(self, 'fix_mode'):
            if self.fix_mode in [ 'all', 'ratio' ]:
                self._total_model = self._initial_model_by_fix_mode(self.fix_mode)
            else:
                log.warning(f'Fix mode "{self.fix_model}" is not support, use "all" mode as default')
                self._total_model = self._initial_model_by_fix_mode('all')
        else:
            log.warning(f'Fix mode is not specified, use "all" mode as default')
            self._total_model = self._initial_model_by_fix_mode('all')

        self._model_built = True

    def set_model(self, params: dict, samples: dict):
        self._initialize()
        self.params  = params
        self.samples = samples

    def save_model(self, path):
        self._initialize()
        utnr.dump_json(self.params, f'{self._pars_dir}/pars_misID.json')

        for key, df in self.samples.items():
            df.to_csv(f'{path}/{key}.csv', index=False)

    def load_model(self):
        self._initialize()
        self.params  = utnr.load_json(f'{self._pars_dir}/pars_misID.json')

        self._samples = {}
        for key in [ 'bpkkk', 'bpkpipi' ]:
            self._samples[key] = pd.read_csv(f'{self._pars_dir}/{key}.csv')
